#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Evaluators;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using System;

namespace HeuristicLab.Problems.DataAnalysis.Regression.Symbolic.Analyzers {
  [Item("OverfittingAnalyzer", "")]
  [StorableClass]
  public sealed class OverfittingAnalyzer : SingleSuccessorOperator, ISymbolicRegressionAnalyzer {
    private const string RandomParameterName = "Random";
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string SymbolicExpressionTreeInterpreterParameterName = "SymbolicExpressionTreeInterpreter";
    private const string ProblemDataParameterName = "ProblemData";
    private const string ValidationSamplesStartParameterName = "SamplesStart";
    private const string ValidationSamplesEndParameterName = "SamplesEnd";
    private const string UpperEstimationLimitParameterName = "UpperEstimationLimit";
    private const string LowerEstimationLimitParameterName = "LowerEstimationLimit";
    private const string EvaluatorParameterName = "Evaluator";
    private const string MaximizationParameterName = "Maximization";
    private const string RelativeNumberOfEvaluatedSamplesParameterName = "RelativeNumberOfEvaluatedSamples";

    #region parameter properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters[RandomParameterName]; }
    }
    public ScopeTreeLookupParameter<SymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ScopeTreeLookupParameter<SymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> ValidationQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["ValidationQuality"]; }
    }
    public IValueLookupParameter<ISymbolicExpressionTreeInterpreter> SymbolicExpressionTreeInterpreterParameter {
      get { return (IValueLookupParameter<ISymbolicExpressionTreeInterpreter>)Parameters[SymbolicExpressionTreeInterpreterParameterName]; }
    }
    public ILookupParameter<ISymbolicRegressionEvaluator> EvaluatorParameter {
      get { return (ILookupParameter<ISymbolicRegressionEvaluator>)Parameters[EvaluatorParameterName]; }
    }
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters[MaximizationParameterName]; }
    }
    public IValueLookupParameter<DataAnalysisProblemData> ProblemDataParameter {
      get { return (IValueLookupParameter<DataAnalysisProblemData>)Parameters[ProblemDataParameterName]; }
    }
    public IValueLookupParameter<IntValue> ValidationSamplesStartParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters[ValidationSamplesStartParameterName]; }
    }
    public IValueLookupParameter<IntValue> ValidationSamplesEndParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters[ValidationSamplesEndParameterName]; }
    }
    public IValueParameter<PercentValue> RelativeNumberOfEvaluatedSamplesParameter {
      get { return (IValueParameter<PercentValue>)Parameters[RelativeNumberOfEvaluatedSamplesParameterName]; }
    }

    public IValueLookupParameter<DoubleValue> UpperEstimationLimitParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters[UpperEstimationLimitParameterName]; }
    }
    public IValueLookupParameter<DoubleValue> LowerEstimationLimitParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters[LowerEstimationLimitParameterName]; }
    }
    public ILookupParameter<PercentValue> RelativeValidationQualityParameter {
      get { return (ILookupParameter<PercentValue>)Parameters["RelativeValidationQuality"]; }
    }
    //public IValueLookupParameter<PercentValue> RelativeValidationQualityLowerLimitParameter {
    //  get { return (IValueLookupParameter<PercentValue>)Parameters["RelativeValidationQualityLowerLimit"]; }
    //}
    //public IValueLookupParameter<PercentValue> RelativeValidationQualityUpperLimitParameter {
    //  get { return (IValueLookupParameter<PercentValue>)Parameters["RelativeValidationQualityUpperLimit"]; }
    //}
    public ILookupParameter<DoubleValue> TrainingValidationQualityCorrelationParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["TrainingValidationCorrelation"]; }
    }
    public IValueLookupParameter<DoubleValue> LowerCorrelationLimitParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["LowerCorrelationLimit"]; }
    }
    public IValueLookupParameter<DoubleValue> UpperCorrelationLimitParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["UpperCorrelationLimit"]; }
    }
    public ILookupParameter<BoolValue> OverfittingParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Overfitting"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ILookupParameter<DoubleValue> InitialTrainingQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["InitialTrainingQuality"]; }
    }
    public ILookupParameter<DoubleMatrix> TrainingAndValidationQualitiesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["TrainingAndValidationQualities"]; }
    }
    public IValueLookupParameter<DoubleValue> PercentileParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["Percentile"]; }
    }
    #endregion
    #region properties
    public IRandom Random {
      get { return RandomParameter.ActualValue; }
    }
    public ItemArray<SymbolicExpressionTree> SymbolicExpressionTree {
      get { return SymbolicExpressionTreeParameter.ActualValue; }
    }
    public ISymbolicExpressionTreeInterpreter SymbolicExpressionTreeInterpreter {
      get { return SymbolicExpressionTreeInterpreterParameter.ActualValue; }
    }
    public ISymbolicRegressionEvaluator Evaluator {
      get { return EvaluatorParameter.ActualValue; }
    }
    public BoolValue Maximization {
      get { return MaximizationParameter.ActualValue; }
    }
    public DataAnalysisProblemData ProblemData {
      get { return ProblemDataParameter.ActualValue; }
    }
    public IntValue ValidiationSamplesStart {
      get { return ValidationSamplesStartParameter.ActualValue; }
    }
    public IntValue ValidationSamplesEnd {
      get { return ValidationSamplesEndParameter.ActualValue; }
    }
    public PercentValue RelativeNumberOfEvaluatedSamples {
      get { return RelativeNumberOfEvaluatedSamplesParameter.Value; }
    }

    public DoubleValue UpperEstimationLimit {
      get { return UpperEstimationLimitParameter.ActualValue; }
    }
    public DoubleValue LowerEstimationLimit {
      get { return LowerEstimationLimitParameter.ActualValue; }
    }
    #endregion

    public OverfittingAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>(RandomParameterName, "The random generator to use."));
      Parameters.Add(new LookupParameter<ISymbolicRegressionEvaluator>(EvaluatorParameterName, "The evaluator which should be used to evaluate the solution on the validation set."));
      Parameters.Add(new ScopeTreeLookupParameter<SymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ValidationQuality"));
      Parameters.Add(new LookupParameter<BoolValue>(MaximizationParameterName, "The direction of optimization."));
      Parameters.Add(new ValueLookupParameter<ISymbolicExpressionTreeInterpreter>(SymbolicExpressionTreeInterpreterParameterName, "The interpreter that should be used for the analysis of symbolic expression trees."));
      Parameters.Add(new ValueLookupParameter<DataAnalysisProblemData>(ProblemDataParameterName, "The problem data for which the symbolic expression tree is a solution."));
      Parameters.Add(new ValueLookupParameter<IntValue>(ValidationSamplesStartParameterName, "The first index of the validation partition of the data set."));
      Parameters.Add(new ValueLookupParameter<IntValue>(ValidationSamplesEndParameterName, "The last index of the validation partition of the data set."));
      Parameters.Add(new ValueParameter<PercentValue>(RelativeNumberOfEvaluatedSamplesParameterName, "The relative number of samples of the dataset partition, which should be randomly chosen for evaluation between the start and end index.", new PercentValue(1)));
      Parameters.Add(new ValueLookupParameter<DoubleValue>(UpperEstimationLimitParameterName, "The upper estimation limit that was set for the evaluation of the symbolic expression trees."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>(LowerEstimationLimitParameterName, "The lower estimation limit that was set for the evaluation of the symbolic expression trees."));
      Parameters.Add(new LookupParameter<PercentValue>("RelativeValidationQuality"));
      //Parameters.Add(new ValueLookupParameter<PercentValue>("RelativeValidationQualityUpperLimit", new PercentValue(0.05)));
      //Parameters.Add(new ValueLookupParameter<PercentValue>("RelativeValidationQualityLowerLimit", new PercentValue(-0.05)));
      Parameters.Add(new LookupParameter<DoubleValue>("TrainingValidationCorrelation"));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("LowerCorrelationLimit", new DoubleValue(0.65)));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("UpperCorrelationLimit", new DoubleValue(0.75)));
      Parameters.Add(new LookupParameter<BoolValue>("Overfitting"));
      Parameters.Add(new LookupParameter<ResultCollection>("Results"));
      Parameters.Add(new LookupParameter<DoubleValue>("InitialTrainingQuality"));
      Parameters.Add(new LookupParameter<DoubleMatrix>("TrainingAndValidationQualities"));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("Percentile", new DoubleValue(1)));

    }

    [StorableConstructor]
    private OverfittingAnalyzer(bool deserializing) : base(deserializing) { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey("InitialTrainingQuality")) {
        Parameters.Add(new LookupParameter<DoubleValue>("InitialTrainingQuality"));
      }
      //if (!Parameters.ContainsKey("RelativeValidationQualityUpperLimit")) {
      //  Parameters.Add(new ValueLookupParameter<PercentValue>("RelativeValidationQualityUpperLimit", new PercentValue(0.05)));
      //}
      //if (!Parameters.ContainsKey("RelativeValidationQualityLowerLimit")) {
      //  Parameters.Add(new ValueLookupParameter<PercentValue>("RelativeValidationQualityLowerLimit", new PercentValue(-0.05)));
      //}
      if (!Parameters.ContainsKey("TrainingAndValidationQualities")) {
        Parameters.Add(new LookupParameter<DoubleMatrix>("TrainingAndValidationQualities"));
      }
      if (!Parameters.ContainsKey("Percentile")) {
        Parameters.Add(new ValueLookupParameter<DoubleValue>("Percentile", new DoubleValue(1)));
      }
      if (!Parameters.ContainsKey("ValidationQuality")) {
        Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ValidationQuality"));
      }
      if (!Parameters.ContainsKey("LowerCorrelationLimit")) {
        Parameters.Add(new ValueLookupParameter<DoubleValue>("LowerCorrelationLimit", new DoubleValue(0.65)));
      }
      if (!Parameters.ContainsKey("UpperCorrelationLimit")) {
        Parameters.Add(new ValueLookupParameter<DoubleValue>("UpperCorrelationLimit", new DoubleValue(0.75)));
      }

    }

    public override IOperation Apply() {
      var trees = SymbolicExpressionTree;
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ItemArray<DoubleValue> validationQualities = ValidationQualityParameter.ActualValue;

      double correlationLimit;
      if (OverfittingParameter.ActualValue != null && OverfittingParameter.ActualValue.Value) {
        // if is already overfitting have to reach the upper limit to switch back to non-overfitting state
        correlationLimit = UpperCorrelationLimitParameter.ActualValue.Value;
      } else {
        // if currently in non-overfitting state have to reach to lower limit to switch to overfitting state
        correlationLimit = LowerCorrelationLimitParameter.ActualValue.Value;
      }
      //string targetVariable = ProblemData.TargetVariable.Value;

      //// select a random subset of rows in the validation set
      //int validationStart = ValidiationSamplesStart.Value;
      //int validationEnd = ValidationSamplesEnd.Value;
      //int seed = Random.Next();
      //int count = (int)((validationEnd - validationStart) * RelativeNumberOfEvaluatedSamples.Value);
      //if (count == 0) count = 1;
      //IEnumerable<int> rows = RandomEnumerable.SampleRandomNumbers(seed, validationStart, validationEnd, count);

      //double upperEstimationLimit = UpperEstimationLimit != null ? UpperEstimationLimit.Value : double.PositiveInfinity;
      //double lowerEstimationLimit = LowerEstimationLimit != null ? LowerEstimationLimit.Value : double.NegativeInfinity;

      //double bestQuality = Maximization.Value ? double.NegativeInfinity : double.PositiveInfinity;
      //SymbolicExpressionTree bestTree = null;

      //List<double> validationQualities = new List<double>();
      //foreach (var tree in trees) {
      //  double quality = Evaluator.Evaluate(SymbolicExpressionTreeInterpreter, tree,
      //    lowerEstimationLimit, upperEstimationLimit,
      //    ProblemData.Dataset, targetVariable,
      //   rows);
      //  validationQualities.Add(quality);
      //  //if ((Maximization.Value && quality > bestQuality) ||
      //  //    (!Maximization.Value && quality < bestQuality)) {
      //  //  bestQuality = quality;
      //  //  bestTree = tree;
      //  //}
      //}

      //if (RelativeValidationQualityParameter.ActualValue == null) {
      // first call initialize the relative quality using the difference between average training and validation quality
      double avgTrainingQuality = qualities.Select(x => x.Value).Average();
      double avgValidationQuality = validationQualities.Select(x => x.Value).Average();

      if (Maximization.Value)
        RelativeValidationQualityParameter.ActualValue = new PercentValue(avgValidationQuality / avgTrainingQuality - 1);
      else {
        RelativeValidationQualityParameter.ActualValue = new PercentValue(avgTrainingQuality / avgValidationQuality - 1);
      }
      //}

      // best first (only for maximization
      var orderedDistinctPairs = (from index in Enumerable.Range(0, qualities.Length)
                                  where qualities[index].Value > 0.0
                                  select new { Training = qualities[index].Value, Validation = validationQualities[index].Value })
                                 .OrderBy(x => -x.Training)
                                 .ToList();

      int n = (int)Math.Round(PercentileParameter.ActualValue.Value * orderedDistinctPairs.Count);

      double[] validationArr = new double[n];
      double[] trainingArr = new double[n];
      //double[,] qualitiesArr = new double[n, 2];
      for (int i = 0; i < n; i++) {
        validationArr[i] = orderedDistinctPairs[i].Validation;
        trainingArr[i] = orderedDistinctPairs[i].Training;

        //qualitiesArr[i, 0] = trainingArr[i];
        //qualitiesArr[i, 1] = validationArr[i];
      }
      double r = alglib.correlation.spearmanrankcorrelation(trainingArr, validationArr, n);
      TrainingValidationQualityCorrelationParameter.ActualValue = new DoubleValue(r);
      if (InitialTrainingQualityParameter.ActualValue == null)
        InitialTrainingQualityParameter.ActualValue = new DoubleValue(avgValidationQuality);
      bool overfitting =
        avgTrainingQuality > InitialTrainingQualityParameter.ActualValue.Value &&  // better on training than in initial generation
        // RelativeValidationQualityParameter.ActualValue.Value < 0.0 && // validation quality is worse than training quality
        r < correlationLimit;


      OverfittingParameter.ActualValue = new BoolValue(overfitting);
      //TrainingAndValidationQualitiesParameter.ActualValue = new DoubleMatrix(qualitiesArr);
      return base.Apply();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void Initialize() { }

    private static void AddValue(DataTable table, double data, string name, string description) {
      DataRow row;
      table.Rows.TryGetValue(name, out row);
      if (row == null) {
        row = new DataRow(name, description);
        row.Values.Add(data);
        table.Rows.Add(row);
      } else {
        row.Values.Add(data);
      }
    }
  }
}
